@@ -35,7 +35,7 @@
 
 module Data.IntervalMap.FingerTree (
     -- * Intervals
-    Interval(..), point,
+    Interval(..), low, high, point,
     -- * Interval maps
     IntervalMap, empty, singleton, insert, union,
     -- * Searching
@@ -45,25 +45,43 @@ module Data.IntervalMap.FingerTree (
 import qualified Data.FingerTree as FT
 import Data.FingerTree (FingerTree, Measured(..), ViewL(..), (<|), (><))
 
+import Prelude hiding (null)
+#if MIN_VERSION_base(4,8,0)
+import qualified Prelude (null)
+#else
 import Control.Applicative ((<$>))
-import Data.Traversable (Traversable(traverse))
 import Data.Foldable (Foldable(foldMap))
 import Data.Monoid
+import Data.Traversable (Traversable(traverse))
+#endif
+#if MIN_VERSION_base(4,9,0)
+import Data.Semigroup
+#endif
+import Data.Foldable (toList)
 
 ----------------------------------
 -- 4.8 Application: interval trees
 ----------------------------------
 
 -- | A closed interval.  The lower bound should be less than or equal
--- to the higher bound.
-data Interval v = Interval { low :: v, high :: v }
+-- to the upper bound.
+data Interval v = Interval v v -- ^ Lower and upper bounds of the interval.
     deriving (Eq, Ord, Show)
 
+-- | Lower bound of the interval
+low :: Interval v -> v
+low (Interval lo _) = lo
+
+-- | Upper bound of the interval
+high :: Interval v -> v
+high (Interval _ hi) = hi
+
 -- | An interval in which the lower and upper bounds are equal.
 point :: v -> Interval v
 point v = Interval v v
 
 data Node v a = Node (Interval v) a
+    deriving (Eq, Ord, Show)
 
 instance Functor (Node v) where
     fmap f (Node i x) = Node i (f x)
@@ -77,19 +95,28 @@ instance Traversable (Node v) where
 -- rightmost interval (including largest lower bound) and largest upper bound.
 data IntInterval v = NoInterval | IntInterval (Interval v) v
 
+#if MIN_VERSION_base(4,9,0)
+instance Ord v => Semigroup (IntInterval v) where
+    (<>) = intervalUnion
+#endif
+
 instance Ord v => Monoid (IntInterval v) where
     mempty = NoInterval
-    NoInterval `mappend` i  = i
-    i `mappend` NoInterval  = i
-    IntInterval _ hi1 `mappend` IntInterval int2 hi2 =
-        IntInterval int2 (max hi1 hi2)
+#if !(MIN_VERSION_base(4,11,0))
+    mappend = intervalUnion
+#endif
+
+intervalUnion :: Ord v => IntInterval v -> IntInterval v -> IntInterval v
+NoInterval `intervalUnion` i  = i
+i `intervalUnion` NoInterval  = i
+IntInterval _ hi1 `intervalUnion` IntInterval int2 hi2 =
+    IntInterval int2 (max hi1 hi2)
+
 
 instance (Ord v) => Measured (IntInterval v) (Node v a) where
     measure (Node i _) = IntInterval i (high i)
 
 -- | Map of closed intervals, possibly with duplicates.
--- The 'Foldable' and 'Traversable' instances process the intervals in
--- lexicographical order.
 newtype IntervalMap v a =
     IntervalMap (FingerTree (IntInterval v) (Node v a))
 -- ordered lexicographically by interval
@@ -97,17 +124,48 @@ newtype IntervalMap v a =
 instance Functor (IntervalMap v) where
     fmap f (IntervalMap t) = IntervalMap (FT.unsafeFmap (fmap f) t)
 
+-- | Values in lexicographical order of intervals.
 instance Foldable (IntervalMap v) where
     foldMap f (IntervalMap t) = foldMap (foldMap f) t
+#if MIN_VERSION_base(4,8,0)
+    null (IntervalMap t) = FT.null t
+#endif
 
+-- | Traverse the intervals in lexicographical order.
 instance Traversable (IntervalMap v) where
     traverse f (IntervalMap t) =
         IntervalMap <$> FT.unsafeTraverse (traverse f) t
 
+instance (Eq v, Eq a) => Eq (IntervalMap v a) where
+    IntervalMap xs == IntervalMap ys = toList xs == toList ys
+
+-- | Lexicographical ordering
+instance (Ord v, Ord a) => Ord (IntervalMap v a) where
+    compare (IntervalMap xs) (IntervalMap ys) = compare (toList xs) (toList ys)
+
+instance (Show v, Show a) => Show (IntervalMap v a) where
+    showsPrec p (IntervalMap ns)
+      | FT.null ns = showString "empty"
+      | otherwise =
+        showParen (p > 0) (showIntervals (toList ns))
+      where
+        showIntervals [] = showString "empty"
+        showIntervals (Node i x:ixs) =
+            showString "insert " . shows i . showChar ' ' . shows x .
+                showString " $ " . showIntervals ixs
+
+#if MIN_VERSION_base(4,9,0)
+-- | 'union'.
+instance (Ord v) => Semigroup (IntervalMap v a) where
+    (<>) = union
+#endif
+
 -- | 'empty' and 'union'.
 instance (Ord v) => Monoid (IntervalMap v a) where
     mempty = empty
+#if !(MIN_VERSION_base(4,11,0))
     mappend = union
+#endif
 
 -- | /O(1)/.  The empty interval map.
 empty :: (Ord v) => IntervalMap v a
